"""
Functions to use with the Factory dependency injection.
"""

from typing import Any, Callable, Tuple, cast

from lilya._internal._module_loading import import_string

from esmerald.exceptions import ImproperlyConfigured


def _lookup(klass: Any, comp: Any, import_path: Any) -> Any:  # pragma: no cover
    """
    Runs a lookup via __import__ and returns the component.
    """
    try:
        return getattr(klass, comp)
    except AttributeError:
        __import__(import_path)
        return getattr(klass, comp)


def _importer(target: Any, attribute: Any) -> Any:  # pragma: no cover
    """
    Gets the attribute from the target.
    """
    components = target.split(".")
    import_path = components.pop(0)
    klass = __import__(import_path)

    for comp in components:
        import_path += ".%s" % comp
        klass = _lookup(klass, comp, import_path)
    return getattr(klass, attribute)


def _get_provider_callable(target: str) -> Any:  # pragma: no cover
    try:
        target, attribute = target.rsplit(".", 1)
    except (TypeError, ValueError, AttributeError):
        raise TypeError(f"Need a valid target to lookup. You supplied: {target!r}") from None

    def getter() -> Any:
        return _importer(target, attribute)

    return getter


def load_provider(provider: str) -> Tuple[Callable, bool]:  # pragma: no cover
    """
    Loads any callable by string import. This will make
    sure that there is no need to have all the imports in one
    file to use the `esmerald.injector.Factory`.

    Example:
        # myapp.daos.py
        from esmerald import AsyncDAOProtocol


        class MyDAO(AsyncDAOProtocol):
            ...

        # myapp.urls.py
        from esmerald import Inject, Factory, Gateway

        route_patterns = [
            Gateway(
                ...,
                dependencies={"my_dao": Inject(Factory("myapp.daos.MyDAO"))}
            )
        ]
    """
    if not isinstance(provider, str):
        raise ImproperlyConfigured(
            "The `provider` should be a string with the format <module>.<file>"
        )

    is_nested: bool = False
    try:
        provider_callable = import_string(provider)
    except ModuleNotFoundError:
        target = _get_provider_callable(provider)
        provider_callable = target
        is_nested = True

    if not callable(provider_callable):
        raise ImproperlyConfigured(
            f"The `provider` specified must be a callable, got {type(provider_callable)} instead."
        )

    return cast(Callable, provider_callable), is_nested



from typing import Any, Dict, Optional

from lilya.types import ASGIApp
from typing_extensions import Protocol, runtime_checkable


@runtime_checkable
class ExtensionProtocol(Protocol):  # pragma: no cover
    def __init__(self, app: Optional["ASGIApp"] = None, **kwargs: Dict[Any, Any]): ...

    def extend(self, **kwargs: Any) -> None: ...



from abc import ABC, abstractmethod
from http.cookies import SimpleCookie
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
    cast,
)

from lilya._internal._message import Address as Address  # noqa: F401
from lilya.datastructures import (
    URL as URL,  # noqa: F401
    DataUpload as LilyaUploadFile,  # noqa
    FormData as FormData,  # noqa: F401
    Header as Header,  # noqa: F401
    QueryParam as QueryParam,  # noqa: F401
    Secret as Secret,  # noqa
    State as State,  # noqa: F401
    URLPath as URLPath,  # noqa: F401
)
from lilya.responses import Response as LilyaResponse  # noqa
from pydantic import BaseModel, ConfigDict, field_validator  # noqa
from pydantic._internal._schema_generation_shared import (
    GetJsonSchemaHandler as GetJsonSchemaHandler,
)
from pydantic.json_schema import JsonSchemaValue as JsonSchemaValue
from pydantic_core.core_schema import (
    CoreSchema,
    PlainValidatorFunctionSchema,
    with_info_plain_validator_function as general_plain_validator_function,
)
from typing_extensions import Literal

from esmerald.background import BackgroundTask, BackgroundTasks  # noqa
from esmerald.enums import MediaType

R = TypeVar("R", bound=LilyaResponse)

if TYPE_CHECKING:  # pragma: no cover
    from esmerald.applications import Esmerald


class UploadFile(LilyaUploadFile):  # pragma: no cover
    """
    Adding pydantic specific functionalitty for parsing.
    """

    @classmethod
    def __get_validators__(cls: Type["UploadFile"]) -> Iterable[Callable[..., Any]]:
        yield cls.validate

    @classmethod
    def validate(cls: Type["UploadFile"], v: Any) -> Any:
        if not isinstance(v, LilyaUploadFile):
            raise ValueError(f"Expected UploadFile, got: {type(v)}")
        return v

    @classmethod
    def _validate(cls, __input_value: Any, _: Any) -> "UploadFile":
        if not isinstance(__input_value, LilyaUploadFile):
            raise ValueError(f"Expected UploadFile, got: {type(__input_value)}")
        return cast(UploadFile, __input_value)

    @classmethod
    def __get_pydantic_json_schema__(
        cls, core_schema: CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        return {"type": "string", "format": "binary"}

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source: Type[Any], handler: Callable[[Any], CoreSchema]
    ) -> CoreSchema:
        return cast(PlainValidatorFunctionSchema, general_plain_validator_function(cls._validate))


class Cookie(BaseModel):
    key: str
    value: Optional[str] = None
    max_age: Optional[int] = None
    expires: Optional[int] = None
    path: str = "/"
    domain: Optional[str] = None
    secure: Optional[bool] = None
    httponly: Optional[bool] = None
    samesite: Literal["lax", "strict", "none"] = "lax"
    description: Optional[str] = None

    def to_header(self, **kwargs: Any) -> str:
        simple_cookie: SimpleCookie = SimpleCookie()
        simple_cookie[self.key] = self.value or ""
        if self.max_age:
            simple_cookie[self.key]["max-age"] = self.max_age
        cookie_dict = self.model_dump()
        for key in ["expires", "path", "domain", "secure", "httponly", "samesite"]:
            if cookie_dict[key] is not None:
                simple_cookie[self.key][key] = cookie_dict[key]
        return simple_cookie.output(**kwargs).strip()


class ResponseContainer(BaseModel, ABC, Generic[R]):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    background: Optional[Union[BackgroundTask, BackgroundTasks]] = None
    headers: Dict[str, Any] = {}
    cookies: List[Cookie] = []

    @abstractmethod
    def to_response(
        self,
        headers: Dict[str, Any],
        media_type: Union["MediaType", str],
        status_code: int,
        app: Type["Esmerald"],
    ) -> R:  # pragma: no cover
        raise NotImplementedError("not implemented")


class ResponseHeader(BaseModel):
    value: Optional[Any] = None

    @field_validator("value")  # type: ignore
    def validate_value(cls, value: Any, values: Dict[str, Any]) -> Any:
        if value is not None:
            return value



from enum import Enum


class StrEnum(str, Enum):
    def __str__(self) -> str:
        return self.value  # type: ignore

    def __repr__(self) -> str:
        return str(self)


class HttpMethod(StrEnum, Enum):
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    PATCH = "PATCH"
    DELETE = "DELETE"
    HEAD = "HEAD"
    OPTIONS = "OPTIONS"
    TRACE = "TRACE"


class MediaType(StrEnum, Enum):
    JSON = "application/json"
    HTML = "text/html"
    TEXT = "text/plain"
    MESSAGE_PACK = "application/x-msgpack"
    TEXT_CHARSET = "text/plain; charset=utf-8"
    PNG = "image/png"
    OCTET = "application/octet-stream"


class OpenAPIMediaType(StrEnum, Enum):
    OPENAPI_YAML = "application/vnd.oai.openapi"
    OPENAPI_JSON = "application/vnd.oai.openapi+json"


class EncodingType(StrEnum, Enum):
    JSON = "application/json"
    MULTI_PART = "multipart/form-data"
    URL_ENCODED = "application/x-www-form-urlencoded"


class ScopeType(StrEnum, Enum):
    HTTP = "http"
    WEBSOCKET = "websocket"


class ParamType(StrEnum, Enum):
    PATH = "path"
    QUERY = "query"
    COOKIE = "cookie"
    HEADER = "header"



from contextlib import suppress
from json import JSONDecodeError, loads
from typing import TYPE_CHECKING, Any, Dict, List, get_args, get_origin

from lilya.datastructures import DataUpload as LilyaUploadFile
from pydantic import BaseModel, ConfigDict
from pydantic.fields import FieldInfo

from esmerald.datastructures import UploadFile
from esmerald.enums import EncodingType

if TYPE_CHECKING:  # pragma: no cover
    from lilya.datastructures import FormData


class HashableBaseModel(BaseModel):  # pragma: no cover
    """
    Pydantic BaseModel by default doesn't handle with hashable types the same way
    a python object would and therefore there are types that are mutable (list, set)
    not hashable and those need to be handled properly.

    HashableBaseModel handles those corner cases.
    """

    def __hash__(self) -> int:
        values: Dict[str, Any] = {}
        for key, value in self.__dict__.items():
            values[key] = None
            if isinstance(value, (list, set)):
                values[key] = tuple(value)
            else:
                values[key] = value
        return hash((type(self),) + tuple(values))


class ArbitraryHashableBaseModel(HashableBaseModel):
    """
    Same as HashableBaseModel but allowing arbitrary values
    """

    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)


class BaseModelExtra(BaseModel):
    """
    BaseModel that allows extra to be passed.
    """

    model_config = ConfigDict(extra="allow")


class ArbitraryBaseModel(BaseModel):
    """
    ArbitraryBaseModel that allows arbitrary_types_allowed to be passed.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)


class ArbitraryExtraBaseModel(BaseModel):
    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)


def flatten(values: List[Any]) -> List[Any]:
    """
    Flattens a list
    """
    flattened = []

    for value in values:
        if isinstance(value, list):
            flattened.extend(flatten(value))
        else:
            flattened.append(value)
    return flattened


def parse_form_data(
    media_type: "EncodingType", form_data: "FormData", field: "FieldInfo"
) -> Any:  # pragma: no cover
    """
    Converts, parses and transforms a multidict into a dict and tries to load them all into
    json.
    """
    values_dict: Dict[str, Any] = {}
    for key, value in form_data.multi_items():
        if not isinstance(value, LilyaUploadFile):
            with suppress(JSONDecodeError):
                value = loads(value)
        value_in_dict = values_dict.get(key)
        if isinstance(value_in_dict, list):
            values_dict[key].append(value)
        elif value_in_dict:
            values_dict[key] = [value_in_dict, value]
        else:
            values_dict[key] = value

    if media_type == EncodingType.MULTI_PART:
        if get_origin(field.annotation) is list:
            values = list(values_dict.values())
            return flatten(values=values)
        if field.annotation in (LilyaUploadFile, UploadFile) and values_dict:
            return list(values_dict.values())[0]

        # Check the arguments if there is any MULTI_PART in a possible Union with UploadFile
        # and a None (Optional).
        for arg in get_args(field.annotation):
            if issubclass(arg, (LilyaUploadFile, UploadFile)) and values_dict:
                return list(values_dict.values())[0]

    return values_dict if values_dict else None



from typing import Any, Literal, Optional

from esmerald.openapi.enums import APIKeyIn, SecuritySchemeType
from esmerald.openapi.security.base import HTTPBase


class APIKeyInQuery(HTTPBase):
    def __init__(
        self,
        *,
        type_: Literal[
            "apiKey", "http", "mutualTLS", "oauth2", "openIdConnect"
        ] = SecuritySchemeType.apiKey.value,
        scheme_name: Optional[str] = None,
        description: Optional[str] = None,
        in_: Optional[Literal["query", "header", "cookie"]] = APIKeyIn.query.value,
        name: Optional[str] = None,
        **kwargs: Any,
    ):
        super().__init__(
            type_=type_,
            description=description,
            name=name,
            in_=in_,
            scheme_name=scheme_name or self.__class__.__name__,
            **kwargs,
        )


class APIKeyInHeader(HTTPBase):
    def __init__(
        self,
        *,
        type_: Literal[
            "apiKey", "http", "mutualTLS", "oauth2", "openIdConnect"
        ] = SecuritySchemeType.apiKey.value,
        scheme_name: Optional[str] = None,
        description: Optional[str] = None,
        in_: Optional[Literal["query", "header", "cookie"]] = APIKeyIn.header.value,
        name: Optional[str] = None,
        **kwargs: Any,
    ):
        super().__init__(
            type_=type_,
            description=description,
            name=name,
            in_=in_,
            scheme_name=scheme_name or self.__class__.__name__,
            **kwargs,
        )


class APIKeyInCookie(HTTPBase):
    def __init__(
        self,
        *,
        type_: Literal[
            "apiKey", "http", "mutualTLS", "oauth2", "openIdConnect"
        ] = SecuritySchemeType.apiKey.value,
        scheme_name: Optional[str] = None,
        description: Optional[str] = None,
        in_: Optional[Literal["query", "header", "cookie"]] = APIKeyIn.cookie.value,
        name: Optional[str] = None,
        **kwargs: Any,
    ):
        super().__init__(
            type_=type_,
            description=description,
            name=name,
            in_=in_,
            scheme_name=scheme_name or self.__class__.__name__,
            **kwargs,
        )



from typing import ClassVar, List

from typing_extensions import Annotated, Doc

from esmerald.routing.apis.views import ListView, SimpleAPIView


class GenericMixin:
    __is_generic__: ClassVar[bool] = True


class CreateAPIView(GenericMixin, SimpleAPIView):
    """
    This class has the same available parameters as the parent,
    `View` but subclassing the `SimpleAPIView`.

    ```python
    from esmerald.routing.apis.generics import CreateAPIView
    ```

    **Example**

    ```python
    from esmerald import patch, post, put
    from esmerald.routing.apis.generics import CreateAPIView


    class UserAPI(CreateAPIView):
        '''
        CreateAPIView only allows the `post`, `put` and `patch`
        to be used by default.
        '''

        @post()
        async def post(self) -> str:
            ...

        @put()
        async def put(self) -> str:
            ...

        @patch()
        async def patch(self) -> str:
        ...
    ```
    """

    http_allowed_methods: Annotated[
        List[str],
        Doc(
            """
            Allowed methods for the given base class.
            """
        ),
    ] = ["post", "put", "patch"]


class DeleteAPIView(GenericMixin, SimpleAPIView):
    """
    This class has the same available parameters as the parent,
    `View` but subclassing the `SimpleAPIView`.

    ```python
    from esmerald.routing.apis.generics import CreateAPIView
    ```

    **Example**

    ```python
    from esmerald import delete
    from esmerald.routing.apis.generics import DeleteAPIView


    class UserAPI(DeleteAPIView):
        '''
        DeleteAPIView only allows the `delete` to be used by default.
        '''

        @delete()
        async def delete(self) -> None:
            ...
    ```
    """

    http_allowed_methods: Annotated[
        List[str],
        Doc(
            """
            Allowed methods for the given base class.
            """
        ),
    ] = ["delete"]


class ReadAPIView(GenericMixin, SimpleAPIView):
    """
    This class has the same available parameters as the parent,
    `View` but subclassing the `SimpleAPIView`.

    ```python
    from esmerald.routing.apis.generics import ReadAPIView
    ```

    **Example**

    ```python
    from esmerald import get
    from esmerald.routing.apis.generics import ReadAPIView


    class UserAPI(ReadAPIView):
        '''
        ReadAPIView only allows the `get` to be used by default..
        '''

        @get()
        async def get(self) -> None:
            ...
    ```
    """

    http_allowed_methods: Annotated[
        List[str],
        Doc(
            """
            Allowed methods for the given base class.
            """
        ),
    ] = ["get"]


class ListAPIView(GenericMixin, ListView):
    """
    Only allows the return to be lists.
    """

    ...



from .base import (
    Error,
    FileResponse,
    HTMLResponse,
    LilyaResponse,
    PlainText,
    Response,
    StreamingResponse,
)
from .encoders import ORJSONResponse as JSONResponse
from .template import TemplateResponse

__all__ = [
    "Error",
    "FileResponse",
    "HTMLResponse",
    "JSONResponse",
    "PlainText",
    "PlainText",
    "Response",
    "LilyaResponse",
    "StreamingResponse",
    "TemplateResponse",
]



from typing import TYPE_CHECKING, Any, Dict, Type, Union

from lilya.responses import RedirectResponse  # noqa
from typing_extensions import Annotated, Doc

from esmerald.datastructures.base import ResponseContainer  # noqa
from esmerald.enums import MediaType

if TYPE_CHECKING:  # pragma: no cover
    from esmerald.applications import Esmerald


class Redirect(ResponseContainer[RedirectResponse]):
    path: Annotated[
        str,
        Doc(
            """
            The url path to redirect.
            This should be in format of `/<str>`.

            Example: `/redirect-page`.
            """
        ),
    ]

    def to_response(
        self,
        headers: Dict[str, Any],
        media_type: Union["MediaType", str],
        status_code: int,
        app: Type["Esmerald"],
    ) -> RedirectResponse:
        return RedirectResponse(
            headers=headers,
            status_code=status_code,
            url=self.path,
            background=self.background,
        )



from typing import Dict, Optional

from pydantic import BaseModel, ConfigDict


class Discriminator(BaseModel):
    """When request bodies or response payloads may be one of a number of
    different schemas, a `discriminator` object can be used to aid in
    serialization, deserialization, and validation.

    The discriminator is a specific object in a schema which is used to inform the consumer of the specification
    of an alternative schema based on the value associated with it.

    When using the discriminator, _inline_ schemas will not be considered.
    """

    model_config = ConfigDict(
        extra="ignore",
        json_schema_extra={
            "examples": [
                {
                    "propertyName": "petType",
                    "mapping": {
                        "dog": "#/components/schemas/Dog",
                        "monster": "https://gigantic-server.com/schemas/Monster/schema.json",
                    },
                }
            ]
        },
    )

    propertyName: str
    """
    **REQUIRED**. The name of the property in the payload that will hold the discriminator value.
    """

    mapping: Optional[Dict[str, str]] = None
    """
    An object to hold mappings between payload values and schema names or references.
    """



from typing import Any

from esmerald.exceptions import EsmeraldAPIException


class DirectiveError(EsmeraldAPIException):
    """
    Exception indicating a problem while executing a directive.
    """

    def __init__(self, detail: str, returncode: int = 1, **kwargs: Any):
        self.returncode = returncode
        super().__init__(detail=detail, **kwargs)



from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generic,
    NoReturn,
    Optional,
    Sequence,
    TypeVar,
    Union,
    cast,
)

from lilya import status
from lilya.responses import (
    Error as Error,
    FileResponse as FileResponse,  # noqa
    HTMLResponse as HTMLResponse,  # noqa
    JSONResponse as JSONResponse,  # noqa
    Ok as Ok,
    PlainText as PlainText,  # noqa
    RedirectResponse as RedirectResponse,  # noqa
    Response as LilyaResponse,  # noqa
    StreamingResponse as StreamingResponse,  # noqa
)
from orjson import OPT_OMIT_MICROSECONDS, OPT_SERIALIZE_NUMPY, dumps
from typing_extensions import Annotated, Doc

from esmerald.encoders import Encoder, json_encoder
from esmerald.enums import MediaType
from esmerald.exceptions import ImproperlyConfigured

PlainTextResponse = PlainText

if TYPE_CHECKING:  # pragma: no cover
    from esmerald.background import BackgroundTask, BackgroundTasks
    from esmerald.types import ResponseCookies

T = TypeVar("T")


class Response(LilyaResponse, Generic[T]):
    """
    Default `Response` object from Esmerald where it can be as the
    return annotation of a [handler](https://esmerald.dev/routing/handlers/).

    Esmerakd automatically will understand what time of response is going to be
    used and parse all the details automatically.

    **Example**

    ```python
    from pydantic import BaseModel

    from esmerald import Esmerald, Gateway, Response, get
    from esmerald.datastructures import Cookie


    @get(path="/me")
    async def home() -> Response:
        return Response(
            Item(id=1, sku="sku1238"),
            headers={"SKY-HEADER": "sku-xyz"},
            cookies=[Cookie(key="sku", value="a-value")],
        )


    Esmerald(routes=[Gateway(handler=home)])
    ```
    """

    def __init__(
        self,
        content: Annotated[
            Any,
            Doc(
                """
                Any content being sent to the response.
                """
            ),
        ],
        *,
        status_code: Annotated[
            int,
            Doc(
                """
                The response status code.
                """
            ),
        ] = status.HTTP_200_OK,
        media_type: Annotated[
            Optional[Union[MediaType, str]],
            Doc(
                """
                The media type used in the response.
                """
            ),
        ] = MediaType.JSON,
        background: Annotated[
            Optional[Union["BackgroundTask", "BackgroundTasks"]],
            Doc(
                """
                Any instance of a [BackgroundTask or BackgroundTasks](https://esmerald.dev/background-tasks/).
                """
            ),
        ] = None,
        headers: Annotated[
            Optional[Dict[str, Any]],
            Doc(
                """
                Any additional headers being passed to the response.
                """
            ),
        ] = None,
        cookies: Annotated[
            Optional["ResponseCookies"],
            Doc(
                """
                A sequence of `esmerald.datastructures.Cookie` objects.

                Read more about the [Cookies](https://esmerald.dev/extras/cookie-fields/?h=responsecook#cookie-from-response-cookies).

                **Example**

                ```python
                from esmerald import Response
                from esmerald.datastructures import Cookie

                response_cookies=[
                    Cookie(
                        key="csrf",
                        value="CIwNZNlR4XbisJF39I8yWnWX9wX4WFoz",
                        max_age=3000,
                        httponly=True,
                    )
                ]

                Response(response_cookies=response_cookies)
                ```
                """
            ),
        ] = None,
        encoders: Annotated[
            Union[Sequence[Encoder], None],
            Doc(
                """
                A sequence of `esmerald.encoders.Encoder` type of objects to be used
                by the response object directly.

                **Example**

                ```python
                from esmerald import Response
                from esmerald.encoders import PydanticEncoder, MsgSpecEncoder

                response_cookies=[
                    encoders=[PydanticEncoder, MsgSpecEncoder]
                ]

                Response(response_cookies=response_cookies)
                ```
                """
            ),
        ] = None,
    ) -> None:
        super().__init__(
            content=content,
            status_code=status_code,
            headers=headers or {},
            media_type=media_type,
            background=cast("BackgroundTask", background),
            encoders=encoders,
        )
        self.cookies = cookies or []

    @staticmethod
    def transform(value: Any) -> Dict[str, Any]:
        """
        The transformation of the data being returned.

        Supports all the default encoders from Lilya and custom from Esmerald.
        """
        return cast(Dict[str, Any], json_encoder(value))

    def make_response(self, content: Any) -> Union[bytes, str]:
        try:
            if (
                content is None
                or content is NoReturn
                and (
                    self.status_code < 100
                    or self.status_code
                    in {status.HTTP_204_NO_CONTENT, status.HTTP_304_NOT_MODIFIED}
                )
            ):
                return b""
            if self.media_type == MediaType.JSON:
                return dumps(
                    content,
                    default=self.transform,
                    option=OPT_SERIALIZE_NUMPY | OPT_OMIT_MICROSECONDS,
                )
            return super().make_response(content)
        except (AttributeError, ValueError, TypeError) as e:  # pragma: no cover
            raise ImproperlyConfigured("Unable to serialize response content") from e



"""
Functions to use with the Router.
"""

from importlib import import_module
from typing import TYPE_CHECKING, Any, Optional, Sequence, Union

from esmerald.exceptions import ImproperlyConfigured

if TYPE_CHECKING:  # pragma: no cover
    from esmerald.routing.gateways import Gateway, WebSocketGateway
    from esmerald.routing.router import Include

DEFAULT_PATTERN = "route_patterns"


def include(
    arg: Any, pattern: Optional[str] = DEFAULT_PATTERN
) -> Sequence[Union["Gateway", "WebSocketGateway", "Include"]]:
    """Simple retrieve functionality to make it easier to include
    routes in the urls. Example, nested routes.

    Example:

        # myapp.routes.py

        from esmerald import Router
        from .views import MyView

        router_patterns = [
            Router(path='/my-iew', routes=[MyView])
        ]

        # routers.py
        from esmerald import Router
        from esmerald.conf.urls import include

        router_patterns = [
            Router(path='/', routes=[include('myapp.routes')])
        ]

    If a `routes.py ` doesn't contain a default `router_patterns` that can be
    also specified dynamically.

    Example:

        # myapp.routes.py

        from esmerald import Router
        from .views import MyView

        myapp_urls = [
            Router(path='/my-iew', route_handlers=[MyView])
        ]

        # routers.py
        from esmerald import Router
        from esmerald.conf.urls import include

        router_patterns = [
            Router(path='/', routes=[include('myapp.routes', pattern='myapp_urls')])
        ]
    """
    if not isinstance(arg, str):
        raise ImproperlyConfigured("The value should be a string with the format <module>.<file>")

    router_conf_module = import_module(arg)
    pattern = pattern or DEFAULT_PATTERN
    patterns = getattr(router_conf_module, pattern, None)
    if not patterns:
        raise ImproperlyConfigured(
            f"There is no pattern {pattern} found in {arg}. "
            "Are you sure you configured it correctly?"
        )

    if not isinstance(patterns, list):
        raise ImproperlyConfigured(f"{patterns} should be a list and not {type(patterns)}.")

    return patterns



from typing import Any, Dict, List, Type, Union

from pydantic import BaseModel, ConfigDict, DirectoryPath
from typing_extensions import Annotated, Doc

from esmerald.protocols.template import TemplateEngineProtocol
from esmerald.template.jinja import JinjaTemplateEngine


class TemplateConfig(BaseModel):
    """
    An instance of [TemplateConfig](https://esmerald.dev/configurations/template/).

    This configuration is a simple set of configurations that when passed enables the template engine.

    **Example**

    ```python
    from esmerald import Esmerald
    from esmerald.config.template import TemplateConfig
    from esmerald.template.jinja import JinjaTemplateEngine

    template_config = TemplateConfig(
        directory=Path("templates"),
        engine=JinjaTemplateEngine,
    )

    app = Esmerald(template_config=template_config)
    ```
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    engine: Annotated[
        Type[TemplateEngineProtocol],
        Doc(
            """
            The template engine to be used.
            """
        ),
    ] = JinjaTemplateEngine

    directory: Annotated[
        Union[DirectoryPath, List[DirectoryPath]],
        Doc(
            """
            The directory for the templates in the format of a path like.

            Example: `/templates`.
            """
        ),
    ]
    env: Annotated[
        Union[Any, None],
        Doc(
            """
            The environment for the template engine.
            This env is **only used for jinja2** templates and its ignored for other template engines.

            The `env` is a `jinja2.Environment` instance.
            """
        ),
    ] = None
    env_options: Annotated[
        Union[Dict[Any, Any], None],
        Doc(
            """
            The options for the template engine. These options are passed to the template engine.

            In the case of the `jinja2` template engine, these options are passed to the `jinja2.Environment` instance and will populate the `env_options` parameter.

            This is currently only used for the `jinja2` template engine and nothing else.
            """
        ),
    ] = {}



from typing import Optional

from lilya._internal._connection import Connection as Connection  # noqa: F401
from lilya.websockets import (
    WebSocket as WebSocket,  # noqa
    WebSocketClose as WebSocketClose,  # noqa
    WebSocketDisconnect as LilyaWebSocketDisconnect,  # noqa
    WebSocketState as WebSocketState,  # noqa
)


class WebSocketDisconnect(LilyaWebSocketDisconnect):
    """Esmerald WebSocketDisconnect"""

    def __init__(self, code: int = 1000, reason: Optional[str] = None) -> None:
        super().__init__(code, reason)  # pragma: no cover



from datetime import datetime, timezone


def convert_time(date: datetime) -> datetime:
    """
    Converts the date into a UTC compatible format.
    """
    if date.tzinfo is not None:
        date.astimezone(timezone.utc)
    return date.replace(microsecond=0)



from lilya import status

from esmerald.enums import HttpMethod

RESERVED_KWARGS = {
    "state",
    "headers",
    "cookies",
    "request",
    "context",
    "socket",
    "data",
    "query",
    "payload",
}
REQUIRED = "required"
IS_DEPENDENCY = "is_dependency"
SKIP_VALIDATION = "skip_validation"

REDIRECT_STATUS_CODES = {
    status.HTTP_301_MOVED_PERMANENTLY,
    status.HTTP_302_FOUND,
    status.HTTP_303_SEE_OTHER,
    status.HTTP_307_TEMPORARY_REDIRECT,
    status.HTTP_308_PERMANENT_REDIRECT,
}

SOCKET = "socket"
DATA = "data"
PAYLOAD = "payload"
REQUEST = "request"
CONTEXT = "context"

AVAILABLE_METHODS = [
    HttpMethod.GET,
    HttpMethod.HEAD,
    HttpMethod.PATCH,
    HttpMethod.PUT,
    HttpMethod.POST,
    HttpMethod.DELETE,
    HttpMethod.OPTIONS,
    HttpMethod.TRACE,
]
